from .pmf import PMF, PMFCollection


class Modifier(object):
    """
    Base class for Modifiers. 'priority' is used for order of opperations and
    for re-rolls it determines the which re-roll is applied.
    e.g. full re-roll > re-roll failed > re-roll 1s
    """
    priority = 0
    def __init__(self, *args, **kwargs):
        pass

    def modify_dice(self, col, thresh=None,  mod_thresh=None):
        return col

    def modify_re_roll(self, col, thresh=None, mod_thresh=None):
        return col

    def modify_threshold(self, thresh):
        return thresh

    def modify_save(self, save):
        return save

    def modify_ap(self, ap):
        return ap

    def modify_invuln(self, invuln):
        return invuln

    def mod_extra_hit(self):
        return None

    def extra_hit(self):
        return None

    def mod_extra_shot(self):
        return None

    def extra_shot(self):
        return None

    def mod_extra_wound(self):
        return None

    def extra_wound(self):
        return None

    def mod_mortal_wound(self):
        return None

    def mortal_wound(self):
        return None

    def modify_drones(self):
        return False, 7, 7

    def modify_self_wounds(self):
        return 0


class MinimumValue(Modifier):
    def __init__(self, min_val):
        self.min_val = min_val

    def modify_dice(self, col, thresh=None,  mod_thresh=None):
        return col.map(lambda x: x.min(self.min_val))


class ExplodingDice(Modifier):
    def __init__(self, thresh, value):
        self.thresh = max(0, min(7, thresh))
        self.value = value

    def _pmf_collection(self):
        return PMFCollection([PMF.static(0)] * self.thresh + [PMF.static(self.value)] * (7 - self.thresh))


class ModExtraHit(ExplodingDice):
    def mod_extra_hit(self):
        return self._pmf_collection()


class ExtraHit(ExplodingDice):
    def extra_hit(self):
        return self._pmf_collection()


class ModExtraAttack(ExplodingDice):
    def mod_extra_shot(self):
        return self._pmf_collection()


class ExtraShot(ExplodingDice):
    def extra_shot(self):
        return self._pmf_collection()


class ModExtraWound(ExplodingDice):
    def mod_extra_wound(self):
        return self._pmf_collection()


class ExtraWound(ExplodingDice):
    def extra_wound(self):
        return self._pmf_collection()


class GenerateMortalWound(ExplodingDice):
    def mortal_wound(self):
        return self._pmf_collection()


class ModGenerateMortalWound(ExplodingDice):
    def mod_mortal_wound(self):
        return self._pmf_collection()


class Haywire(ExplodingDice):
    def mod_mortal_wound(self):
        col = [PMF.static(0), PMF.static(0), PMF.static(0), PMF.static(0), PMF.static(1), PMF.static(1), PMF.dn(3)]
        return PMFCollection(col)


class ReRollOnes(Modifier):
    priority = 1
    def modify_re_roll(self, col, thresh=None, mod_thresh=None):
        return col.map(lambda x: x.re_roll_value(1))


class ReRollFailed(Modifier):
    priority = 99
    def modify_re_roll(self, col, thresh=None, mod_thresh=None):
        rr_thresh = min(thresh, mod_thresh)
        return col.map(lambda x: x.re_roll_less_than(rr_thresh))


class ReRollOneDice(Modifier):
    def modify_re_roll(self, col, thresh=None, mod_thresh=None):
        pmfs = col.pmfs
        if not pmfs:
            return col
        pmfs[0] = pmfs[0].re_roll_less_than(thresh)
        return PMFCollection(pmfs)

class ModReRollOneDice(Modifier):
    def modify_re_roll(self, col, thresh=None, mod_thresh=None):
        pmfs = col.pmfs
        if not pmfs:
            return col
        pmfs[0] = pmfs[0].re_roll_less_than(mod_thresh)
        return PMFCollection(pmfs)


class ReRollOneDiceVolume(Modifier):
    def modify_re_roll(self, col, thresh=None, mod_thresh=None):
        pmfs = col.pmfs
        if not pmfs:
            return col
        pmfs[0] = pmfs[0].re_roll_less_than(pmfs[0].mean())
        return PMFCollection(pmfs)


class ReRollAll(Modifier):
    priority = 100
    def modify_re_roll(self, col, thresh=None, mod_thresh=None):
        return col.map(lambda x: x.re_roll_less_than(mod_thresh))


class ReRollLessThanExpectedValue(Modifier):
    priority = 98
    def modify_re_roll(self, col, thresh=None, mod_thresh=None):
        return col.map(lambda x: x.re_roll_less_than(x.mean()))


class Melta(Modifier):
    def modify_dice(self, col, thresh=None, mod_thresh=None):
        return col.map(lambda x: x.melta())


class AddNTo(Modifier):
    def __init__(self, value=0, *args, **kwargs):
        self.n = value
        self.priority = self.n


class AddNToThreshold(AddNTo):
    def modify_threshold(self, thresh):
        return thresh - self.n


class AddNToAP(AddNTo):
    def modify_ap(self, ap):
        return ap - self.n


class AddNToSave(AddNTo):
    def modify_save(self, save):
        return save - self.n


class AddNToInvuln(AddNTo):
    def modify_invuln(self, invuln):
        return invuln - self.n


class AddNToVolume(AddNTo):
    def modify_dice(self, col, thresh=None, mod_thresh=None):
        return col.map(lambda x: x.roll(self.n))


class AddD6(AddNTo):
    def modify_dice(self, col, thresh=None, mod_thresh=None):
        return PMFCollection(col.pmfs+[PMF.dn(6)])


class AddD3(AddNTo):
    def modify_dice(self, col, thresh=None, mod_thresh=None):
        return PMFCollection(col.pmfs+[PMF.dn(3)])


class SetToN(Modifier):
    def __init__(self, value=0, *args, **kwargs):
        self.n = value


class SetThresholdToN(SetToN):
    def modify_threshold(self, thresh):
        return self.n


class SetAPToN(SetToN):
    def modify_ap(self, ap):
        return self.n


class SetSaveToN(SetToN):
    def modify_save(self, save):
        return self.n


class SetInvulnToN(SetToN):
    def modify_invuln(self, invuln):
        return self.n


class IgnoreAP(Modifier):
    def __init__(self, value=0, *args, **kwargs):
        self.ap = value

    def modify_ap(self, ap):
        return 0 if ap <= self.ap else ap


class IgnoreInvuln(Modifier):
    def modify_invuln(self, ap):
        return 7


class HalfDamage(Modifier):
    def modify_dice(self, col, thresh=None, mod_thresh=None):
        return col.map(lambda x: x.div_min_one(2))


class ShieldDrone(Modifier):
    def modify_drones(self):
        return True, 2, 5


class NormalDrone(Modifier):
    def modify_drones(self):
        return True, 2, 7


class Overheat(Modifier):
    def modify_self_wounds(self):
        return 2


class ModifierCollection(object):
    """
    Used to keep track of any modifiers to the attack
    """
    def __init__(self, shot_mods=None, hit_mods=None, wound_mods=None, pen_mods=None,
        fnp_mods=None, damage_mods=None):

        self.shot_mods = self._sort_priority(shot_mods or [])
        self.hit_mods = self._sort_priority(hit_mods or [])
        self.wound_mods = self._sort_priority(wound_mods or [])
        self.pen_mods = self._sort_priority(pen_mods or [])
        self.fnp_mods = self._sort_priority(fnp_mods or [])
        self.damage_mods = self._sort_priority(damage_mods or [])

    def __add__(self, other):
        return ModifierCollection(
            shot_mods=self.shot_mods+other.shot_mods,
            hit_mods=self.hit_mods+other.hit_mods,
            wound_mods=self.wound_mods+other.wound_mods,
            pen_mods=self.pen_mods+other.pen_mods,
            fnp_mods=self.fnp_mods+other.fnp_mods,
            damage_mods=self.damage_mods+other.damage_mods,
        )

    def _sort_priority(self, mods ):
        return sorted(mods, key=lambda x: x.priority, reverse=True)

    def _mod_dice(self, col, mods, thresh=None, mod_thresh=None):
        """
        Apply dice modifications. Rerolls happen before modifiers.
        """
        for mod in mods:
            col = mod.modify_re_roll(col, thresh, mod_thresh)
        for mod in mods:
            col = mod.modify_dice(col, thresh, mod_thresh)
        return col

    def modify_shot_dice(self, dists):
        """
        Modify the PMF of shot volume the dice. Ususally for re-rolls.
        """
        return self._mod_dice(dists, self.shot_mods)

    def modify_hit_thresh(self, thresh):
        """
        Modify the hit threshold. Important to note the -1 to hit modifiers actually
        are a +1 to the threshold. Similarly +1 to hits are -1 to the threshold.
        """
        if thresh == 1:
            # Handle the case where the weapon is auto hit. No to hit modifiers map
            return thresh
        for mod in self.hit_mods:
            thresh = mod.modify_threshold(thresh)
        return max(thresh, 2) #1's always fail

    def modify_hit_dice(self, dists, thresh, mod_thresh):
        """
        Modify the PMF of hit dice. Ususally for re-rolls.
        """
        return self._mod_dice(dists, self.hit_mods, thresh, mod_thresh)

    def modify_wound_thresh(self, thresh):
        """
        Modify the wound threshold. Important to note the -1 to wound modifiers actually
        are a +1 to the threshold. Similarly +1 are -1 to the threshold.
        """
        for mod in self.wound_mods:
            thresh = mod.modify_threshold(thresh)
        return max(2, thresh) # 1's always fail

    def modify_wound_dice(self, dists, thresh, mod_thresh):
        """
        Modify the PMF of hit dice. Ususally for re-rolls.
        """
        return self._mod_dice(dists, self.wound_mods, thresh, mod_thresh)

    def modify_pen_thresh(self, save, ap, invuln):
        """
        Modify the pen threshold by modifying the save, ap, and invuln
        """
        for mod in self.pen_mods:
            save = mod.modify_save(save)
        for mod in self.pen_mods:
            ap = mod.modify_ap(ap)
        for mod in self.pen_mods:
            invuln = mod.modify_invuln(invuln)
        return min(max(save + ap, 2), max(invuln, 2)) # 1's alwasys fail

    def modify_pen_dice(self, dists, thresh, mod_thresh):
        """
        Modify the PMF of the pen dice. Ususally for re-rolls.
        """
        return self._mod_dice(dists, self.pen_mods, thresh, mod_thresh)

    def modify_drone(self):
        """
        Return if the attack should be modified by saviour protocols
        """
        enabled = False
        thresh = 7
        fnp = 7
        for mod in self.pen_mods:
            e, t, f = mod.modify_drones()

            enabled = enabled or e
            thresh = thresh if thresh < t else t
            fnp = fnp if fnp < f else f

        return enabled, thresh, fnp

    def modify_self_wounds(self):
        """
        Return threshold for self wound
        """
        thresh = 0
        for mod in self.hit_mods:
            mod_thresh = mod.modify_self_wounds()
            thresh = thresh if thresh > mod_thresh else mod_thresh
        return thresh


    def modify_fnp_thresh(self, thresh):
        """
        Modify the fnp threshold. I think some death guard units can do this?
        """
        for mod in self.fnp_mods:
            thresh = mod.modify_save(thresh)
        return max(thresh, 2) # 1's alwasys fail

    def modify_fnp_dice(self, dists, thresh, mod_thresh):
        """
        Modify the PMF of the FNP dice. Ususally for re-rolls.
        """
        return self._mod_dice(dists, self.fnp_mods, thresh, mod_thresh)

    def modify_damage_dice(self, dists):
        """
        Modify the damage dice
        """
        return self._mod_dice(dists, self.damage_mods)

    def sum_generators(self, mod_list, attr_name):
        cols = [getattr(mod, attr_name)() for mod in mod_list]
        cols = [x for x in cols if x]
        return PMFCollection.add_many(cols)

    def get_mod_extra_hit(self):
        """
        Generate extra hits on a modfiable value
        """
        return self.sum_generators(self.hit_mods, 'mod_extra_hit')

    def get_extra_hit(self):
        """
        Generate extra hits on a static value
        """
        return self.sum_generators(self.hit_mods, 'extra_hit')

    def get_mod_extra_shot(self):
        """
        Generate extra shots on a modfiable value
        """
        return self.sum_generators(self.hit_mods, 'mod_extra_shot')

    def get_extra_shot(self):
        """
        Generate extra shots on a static value
        """
        return self.sum_generators(self.hit_mods, 'extra_shot')

    def get_mod_extra_wound(self):
        """
        Generate extra wounds on a modfiable value
        """
        return self.sum_generators(self.wound_mods, 'mod_extra_wound')

    def get_extra_wound(self):
        """
        Generate extra wounds on a static value
        """
        return self.sum_generators(self.wound_mods, 'extra_wound')

    def hit_mod_mortal_wounds(self):
        """
        Generate mortal wounds on a modfiable value
        """
        return self.sum_generators(self.hit_mods, 'mod_mortal_wound')

    def wound_mod_mortal_wounds(self):
        """
        Generate mortal wounds on a modfiable value
        """
        return self.sum_generators(self.wound_mods, 'mod_mortal_wound')

    def hit_mortal_wounds(self):
        """
        Generate mortal wounds on a static value
        """
        return self.sum_generators(self.hit_mods, 'mortal_wound')

    def wound_mortal_wounds(self):
        """
        Generate mortal wounds on a static value
        """
        return self.sum_generators(self.wound_mods, 'mortal_wound')



from .pmf import PMF, PMFCollection


class AttackSequenceResults(object):
    """
    Holds the results of an attack sequence
    """
    def __init__(self, shot_dist, hit_dist, wound_dist, pen_dist, damage_dist, drone_wound, self_wound, mortal_dist):
        self.shot_dist = shot_dist
        self.hit_dist = hit_dist
        self.wound_dist = wound_dist
        self.pen_dist = pen_dist
        self.damage_dist = damage_dist
        self.drone_wound = drone_wound
        self.self_wound = self_wound
        self.mortal_dist = mortal_dist


class Attack(object):
    """
    Generates the probability distribution for damage dealt from an attack sequence
    """
    def __init__(self, weapon, target):
        self.weapon = weapon
        self.target = target

        # Setup the controllers for the individual phases
        self.shots = AttackShots(self)
        self.hits = AttackHits(self)
        self.wounds = AttackWounds(self)
        self.drones = AttackDrones(self)
        self.pens = AttackPens(self)
        self.damage = AttackDamage(self)

    @property
    def modifiers(self):
        # Return a combined list of modifers for both the weapon ans the target
        return self.weapon.modifiers + self.target.modifiers

    def run(self):
        """
        Generate the resulting PMF
        """
        # Keeping track of mortal wounds generated in the hit and wound phase
        mortal_dists = []

        # Generate the PMF for the number of shots
        shot_dist = self.shots.calc_dist()

        # Generate the PMFs for the number of hits, mortal wounds, and self inflicted
        # mortal wounds in the hit phase
        hit_dist, mortal_dist, self_dist = self.hits.calc_dist(shot_dist)
        mortal_dists.append(mortal_dist)

        # Generate the PMFs for the number of wounds, mortal wounds, and self inflicted
        # mortal wounds in the wound phase
        wound_dist, mortal_dist = self.wounds.calc_dist(hit_dist)
        mortal_dists.append(mortal_dist)

        # Generate the PMFs for the number of wounds that penetrate saviour protocols and
        # how many drones die
        drone_pen, drone_wound = self.drones.calc_dist(wound_dist)

        # Generate the PMFs for the number of failed saves in the armor save phase
        pen_dist = self.pens.calc_dist(drone_pen)

        # Generate the PMFs for the danage dealt in the damage calculation phase.
        damage_dist = self.damage.calc_dist(pen_dist)

        return AttackSequenceResults(
            shot_dist,
            hit_dist,
            wound_dist,
            pen_dist,
            damage_dist,
            drone_wound,
            self_dist,
            PMF.convolve_many(mortal_dists),
        )


class AttackSegment(object):
    """
    The base class for the various attack segments
    """
    def __init__(self, attack):
        self.attack = attack
        self._thresh_mod = None

    def calc_dist(self):
        # Stub
        pass

    @property
    def thresh_mod(self):
        # cache the threshold modifier
        if self._thresh_mod is None:
            self._thresh_mod = self._get_thresh_mod()
        return self._thresh_mod

    def _get_thresh_mod(self):
        return 0

    def _calc_exp_dist(self, dice_dists):
        # Calculate the exploding dice distributions
        dist_col = PMFCollection.add_many([
        self._mod_extra_dist().thresh_mod(self.thresh_mod),
        self._extra_dist(),
        ])

        if dist_col:
            return dist_col.mul_col(dice_dists).convolve()
        else:
            return PMF.static(0)

    def _mod_extra_dist(self):
        # Stub
        return PMFCollection.empty()

    def _extra_dist(self):
        # Stub
        return PMFCollection.empty()

    def _calc_mortal_dist(self, dice_dists):
        # Calculate how many mortal wounds are generated
        dist_col = PMFCollection.add_many([
        self._mod_mortal_dist().thresh_mod(self.thresh_mod),
        self._mortal_dist(),
        ])
        if dist_col:
            return dist_col.mul_col(dice_dists).convolve()
        else:
            return PMF.static(0)

    def _mod_mortal_dist(self):
        # Stub
        return PMFCollection.empty()

    def _mortal_dist(self):
        # Stub
        return PMFCollection.empty()


class AttackShots(AttackSegment):
    """
    Generate the PMF for the number of shots
    """
    def calc_dist(self):
        # Apply the shot volume modifiers
        shots_dists = self.attack.modifiers.modify_shot_dice(self.attack.weapon.shots)

        # Convolve the resulting PMFs
        return shots_dists.convolve()


class AttackHits(AttackSegment):
    """
    Simulate the hits phase of the attack
    """
    def calc_dist(self, dist, can_recurse=True):
        hit_dists = []
        exp_dists = []
        mrt_dists = []
        self_dists = []

        # The bare threshold to hit is the weapons to hit value
        thresh = self.attack.weapon.bs

        # Check for any modifers to the to hit roll
        mod_thresh = self.attack.modifiers.modify_hit_thresh(self.attack.weapon.bs)

        # Check for the threshhold for self inflicted wounds
        thresh_self_wounds = self.attack.modifiers.modify_self_wounds()

        # Generate the resulting hit distributon for each value of the shots PMF then
        # multipy by the probability of that value and sum them
        for dice_count, event_prob in enumerate(dist.values):
            # If the probability is zero then no-op
            if event_prob == 0:
                continue

            # Create a dice distribution for 'dice_count' d6 dice
            dice_dists = PMFCollection.mdn(dice_count, 6)

            # Apply modifiers to the dice distribution
            dice_dists = self.attack.modifiers.modify_hit_dice(dice_dists, thresh, mod_thresh)

            # If self inflicted wounds are possible calculate them
            if thresh_self_wounds:
                self_thresh = max(thresh_self_wounds+mod_thresh-thresh, 0)
                self_dist = dice_dists.convert_binomial_less_than(self_thresh).convolve()
            else:
                self_dist = PMF.static(0)


            # Convert the dice dist into a binomial distribution with the modified to hit value.
            # We are bsically flatteing all the individual d6 down into d2 True/False for if it hit.
            hit_dist = dice_dists.convert_binomial(mod_thresh).convolve()

            # Calculate the distirbution for exploding automatic hits. Since extra shots cannot
            # generate more hits return a zero prob if we are recusing
            exp_dist = self._calc_exp_dist(dice_dists) if can_recurse else PMF.static(0)

            # Calculate the ditributons for exploding extra chances to hit. This requires us to calcuate
            # The whole to-hit phase again for the sub hits
            exp_shot_dist, exp_mrt_dist, exp_self_dist = self._calc_exp_shot_dist(dice_dists, can_recurse)

            # Calculate the mortal wounds gnerated
            mrt_dist = self._calc_mortal_dist(dice_dists)

            # Multiply the resulting hit distributions by the probability of this number of dice
            # and append them to the totals
            hit_dists.append(hit_dist * event_prob)
            exp_dists.append(PMF.convolve_many([exp_dist, exp_shot_dist]) * event_prob)
            mrt_dists.append(PMF.convolve_many([mrt_dist, exp_mrt_dist]) * event_prob)
            self_dists.append(PMF.convolve_many([self_dist, exp_self_dist]) * event_prob)

        # Flattening the dists means we are summing the probabilites of each value across all the
        # dists. This is how we find the total probability distribution of the phase.

        # Since the exploding hits happens as a kind of sidecar opperation to the main dice rolls
        # it can be considered an "AND" opperation meaning we can convolve the two distributions
        # together
        combined_hit_dists = PMF.convolve_many([PMF.flatten(hit_dists), PMF.flatten(exp_dists)])
        return (
            combined_hit_dists,
            PMF.flatten(mrt_dists),
            PMF.static(0) if not self_dists else PMF.flatten(self_dists),
        )

    def _get_thresh_mod(self):
        # This is a bit of a hack to get the delta between the threshold and the modified threshold
        return self.attack.modifiers.modify_hit_thresh(6) - 6

    def _calc_exp_shot_dist(self, dice_dists, can_recurse=True):
        # This handles generating extra chances to hit and then calculating the distribution of
        # resulting successful hits

        # If we are alread recusing then return zeros
        if not can_recurse:
            return PMF.static(0), PMF.static(0), PMF.static(0)

        # Since to hit modifers can modify the number of extra chances to hit generate that now
        dist_col = PMFCollection.add_many([
            self._get_mod_extra_shot().thresh_mod(self.thresh_mod),
            self._get_extra_shot(),
        ])

        if dist_col:
            return self.calc_dist(dist_col.mul_col(dice_dists).convolve(), can_recurse=False)
        else:
            return PMF.static(0), PMF.static(0), PMF.static(0)

    def _mod_extra_dist(self):
        return self.attack.modifiers.get_mod_extra_hit()

    def _extra_dist(self):
        return self.attack.modifiers.get_extra_hit()

    def _get_mod_extra_shot(self):
        return self.attack.modifiers.get_mod_extra_shot()

    def _get_extra_shot(self):
        return self.attack.modifiers.get_extra_shot()

    def _mod_mortal_dist(self):
        return self.attack.modifiers.hit_mod_mortal_wounds()

    def _mortal_dist(self):
        return self.attack.modifiers.hit_mortal_wounds()


class AttackWounds(AttackSegment):
    """
    Generate the PMF for the wounds
    """
    def calc_dist(self, dist):
        wnd_dists = []
        exp_dists = []
        mrt_dists = []

        # Calculate the wound threshold from the strength of the weapon and the toughness
        # of the target
        thresh = self._calc_wound_thresh(
            self.attack.weapon.strength,
            self.attack.target.toughness
        )

        # Apply any modifers to the wound threshold
        mod_thresh = self.attack.modifiers.modify_wound_thresh(thresh)

        # Generate the resulting hit distributon for each value of the shots PMF then
        # multipy by the probability of that value and sum them
        for dice_count, event_prob in enumerate(dist.values):
            # If the probability is zero then no-op
            if event_prob == 0:
                continue

            # Create a dice distribution for 'dice_count' d6 dice
            dice_dists = PMFCollection.mdn(dice_count, 6)

            # Apply modifiers to the dice distribution
            dice_dists = self.attack.modifiers.modify_wound_dice(dice_dists, thresh, mod_thresh)

            # Convert the dice dist into a binomial distribution with the modified to wound value.
            # We are bsically flatteing all the individual d6 down into d2 True/False for if it wounded.
            wnd_dist = dice_dists.convert_binomial(mod_thresh).convolve()

            # Calculate exploding auto-wounds
            exp_dist = self._calc_exp_dist(dice_dists)

            # Generate mortal wounds
            mrt_dist = self._calc_mortal_dist(dice_dists)

            wnd_dists.append(wnd_dist * event_prob)
            exp_dists.append(exp_dist * event_prob)
            mrt_dists.append(mrt_dist * event_prob)

        # Flatten the distros down
        return PMF.convolve_many([PMF.flatten(wnd_dists), PMF.flatten(exp_dists)]), PMF.flatten(mrt_dists)

    def _calc_wound_thresh(self, strength, toughness):
        # Generate the wound threshold
        if strength <= toughness/2.0:
            # If the toughness is greater than twice the strength wound on a 6+
            return 6
        elif strength >= toughness*2:
            # If the stength is greater than twice the toughness wound on a 2+
            return 2
        elif toughness > strength:
            # Else if toughness is greater than strength wound on a 5_
            return 5
        elif toughness == strength:
            # Else if they are equal then wound on a 4+
            return 4
        else:
            # If the strength is greater than toughness wound on a 3+
            return 3

    def _get_thresh_mod(self):
        return self.attack.modifiers.modify_wound_thresh(6) - 6

    def _mod_extra_dist(self):
        return self.attack.modifiers.get_mod_extra_wound()

    def _extra_dist(self):
        return self.attack.modifiers.get_extra_wound()

    def _mod_mortal_dist(self):
        return self.attack.modifiers.wound_mod_mortal_wounds()

    def _mortal_dist(self):
        return self.attack.modifiers.wound_mortal_wounds()


class AttackDrones(AttackSegment):
    """
    Generate the PMF for saviour protocols
    """
    def calc_dist(self, dist):
        # Check if saviour protcols are active and if so what the threshold is and the FNP value
        drone_enabled, drone_thresh, drone_fnp = self.attack.modifiers.modify_drone()

        # If the probability is zero then no-op
        if not drone_enabled:
            return dist, PMF.static(0)

        # Calculate how many wounds pass though the savious protocols and how many wounds the drones take
        pens = []
        saves = []
        for dice_count, event_prob in enumerate(dist.values):
            # If the probability is zero then no-op
            if event_prob == 0:
                continue

            # Create a dice distribution for 'dice_count' d6 dice
            dice_dists = PMFCollection.mdn(dice_count, 6)

            # Convert binomial for values less than the threshold for attacks the penetrated
            pen_dist = dice_dists.convert_binomial_less_than(drone_thresh).convolve()

            # Convert binomial for values greater than the threshold for wounds passed off to the
            # drones
            save_dist = dice_dists.convert_binomial(drone_thresh).convolve()

            pens.append(pen_dist * event_prob)
            saves.append(save_dist * event_prob)

        # Flatten the distros down
        return PMF.flatten(pens), self._calc_fnp_dist(PMF.flatten(saves), drone_fnp)

    def _calc_fnp_dist(self, dist, drone_fnp):
        # Check how many of the wounds are removed by the drone FNP
        dists = []
        for dice, event_prob in enumerate(dist.values):
            if event_prob == 0:
                continue
            dice_dists = PMFCollection.mdn(dice, 6)
            binom_dists = dice_dists.convert_binomial_less_than(drone_fnp).convolve()
            dists.append(binom_dists * event_prob)
        return PMF.flatten(dists)


class AttackPens(AttackSegment):
    """
    Generate the PMF for the wounds that penetrate the
    """
    def calc_dist(self, dist):
        # Calculate the final save threshold from the save value, invunerable save, and
        # any save modifiers
        mod_thresh = self.attack.modifiers.modify_pen_thresh(
            self.attack.target.save,
            self.attack.weapon.ap,
            self.attack.target.invuln
        )

        dists = []
        for dice_count, event_prob in enumerate(dist.values):
            # If the probability is zero then no-op
            if event_prob == 0:
                continue

            # Create a dice distribution for 'dice_count' d6 dice
            dice_dists = PMFCollection.mdn(dice_count, 6)
            dice_dists = self.attack.modifiers.modify_pen_dice(dice_dists, mod_thresh, mod_thresh)

            # Convert binomial for values less than the threshold for attacks the penetrated
            pen_dists = dice_dists.convert_binomial_less_than(mod_thresh).convolve()
            dists.append(pen_dists * event_prob)

        # Flatten the distros down
        return PMF.flatten(dists)


class AttackDamage(AttackSegment):
    """
    Generate the PMF for the damage dealt to the target
    """
    def calc_dist(self, dist):
        # Get the distribution of each individual wound dealt
        individual_dist = self._calc_individual_damage_dist()

        damge_dists = []
        for dice, event_prob in enumerate(dist.values):
            # If the probability is zero then no-op
            if event_prob == 0:
                continue
            # Multiply the individual damage distributions by the number dice
            damge_dists.append(PMF.convolve_many([individual_dist] * dice) * event_prob)
        return PMF.flatten(damge_dists)

    def _calc_individual_damage_dist(self):
        # Get the basic damage distribution
        dice_dists = self.attack.weapon.damage

        # Apply modifers to the damage distribution
        mod_dists = self.attack.modifiers.modify_damage_dice(dice_dists)

        # Calculate the damge distribution after the feel no pain has been applied
        fnp_dist = self._calc_fnp_dist(mod_dists.convolve())

        # Clip the damage dist to the wounds characteristic of the target
        return fnp_dist.ceiling(self.attack.target.wounds)

    def _calc_fnp_dist(self, dist):
        dists = []
        mod_thresh = self.attack.modifiers.modify_fnp_thresh(self.attack.target.fnp)
        for dice, event_prob in enumerate(dist.values):
            if event_prob == 0:
                continue
            dice_dists = self.attack.modifiers.modify_fnp_dice(
                PMFCollection.mdn(dice, 6),
                self.attack.target.fnp,
                mod_thresh,
            )
            binom_dists = dice_dists.convert_binomial_less_than(mod_thresh).convolve()
            dists.append(binom_dists * event_prob)
        return PMF.flatten(dists)





from .modifiers import ModifierCollection

class Target(object):
    """
    Holds params of a target
    """
    def __init__(self, toughness, save, invuln, fnp, wounds, modifiers=None):
        self.toughness = toughness
        self.save = save
        self.invuln = invuln
        self.fnp = fnp
        self.wounds = wounds
        self.modifiers = modifiers or ModifierCollection()


import math
from numpy import fft, zeros


class PMF(object):
    """
    Discrete Probability Distribution - Used to keep track of the probability of random discrete events
    """
    def __init__(self, values=None):
        self.values = values if values is not None else []

    def __str__(self):
        return str(self.values)

    def __len__(self):
        return len(self.values)

    def __mul__(self, other):
        return PMF([other*x for x in self.values])

    def __rmul__(self, other):
        return self.__mul__(other)

    def ceiling(self, value):
        """
        Sum the probability of all values >= the ceiling value
        """
        new_values = []
        for i, prob in enumerate(self.values):
            if i <= value:
                new_values.append(prob)
            else:
                new_values[value] += prob
        return PMF(new_values)

    def trim_tail(self, thresh=10**(-4)):
        """
        Trim the long tail off where p < threshold
        """
        in_body = False
        trimmed_dist = []
        for element in self.values[::-1]:
            if in_body or element >= thresh:
                in_body = True
                trimmed_dist.append(element)
        return PMF(trimmed_dist[::-1])

    def cumulative(self):
        """
        Probability of at least each value
        """
        cumu_dist = []
        for i in range(len(self.values)):
            cumu_dist.append(sum(self.values[i:]))
        return PMF(cumu_dist)

    def re_roll_value(self, value):
        """
        Re-roll a specific value (eg re-roll 1's)
        """
        rr_mask = [0.0 if x == value else 1.0 for x in range(len(self))]
        rr_dist = [self.values[value] * x for x in self.values]
        new_dist = []
        for i, prob in enumerate(self.values):
            new_dist.append(rr_mask[i]*prob + rr_dist[i])
        return PMF(new_dist)

    def re_roll_less_than(self, value):
        """
        Re-roll all values below a specific value
        """
        rr_mask = [0.0 if i < value else x for i, x in enumerate(self.values)]
        for i, _ in enumerate(self.values):
            if i < value:
                for j, _ in enumerate(self.values):
                    rr_mask[j] += self.values[i] * self.values[j]
        return PMF(rr_mask)

    def convert_binomial(self, thresh):
        return PMF([sum(self.values[:thresh]), sum(self.values[thresh:])])

    def convert_binomial_less_than(self, thresh):
        return PMF([sum(self.values[thresh:]), sum(self.values[:thresh])])

    def get(self, value):
        if value < 0:
            return 0
        if value > len(self.values):
            return 0
        else:
            return self.values[value]

    def expand_to(self, length):
        return PMF(self.values + [0.0] * max(length - len(self), 0))

    def expected_value(self):
        return sum([x*p for x, p in enumerate(self.values)])

    def add_value(self, n):
        return PMF([0.0] * n + self.values)

    def melta(self):
        return PMF.max_of_two(self, self)

    def roll(self, n):
        if n == 0:
            return self
        elif n > 0:
            return PMF([0.0] * n + self.values)
        else:
            return PMF([sum(self.values[:(-1*n)+1])] + self.values[(-1*n)+1:])

    def div_min_one(self, divisor):
        new_dist = [0.0] * len(self.values)
        for i, value in enumerate(self.values):
            new_dist[math.ceil(i/divisor)] += value
        return PMF(new_dist)

    def rectify_zero(self):
        new_dist = [1.0 - sum(self.values[1:])] + self.values[1:]
        return PMF(new_dist)

    def min(self, min_val):
        return PMF([0.0] * min_val + [sum(self.values[:min_val+1])] + self.values[min_val+1:])

    def mean(self):
        return sum([x*p for x, p in enumerate(self.values)])

    def std(self):
        mean = self.mean()
        exp_mean = sum([(x**2)*p for x, p in enumerate(self.values)])
        return (exp_mean - mean**2)**(0.5)

    @classmethod
    def dn(cls, n):
        return PMF([0.0] + [1/n] * n)

    @classmethod
    def static(cls, n):
        return PMF([0.0] * n + [1.0])

    @classmethod
    def convolve_many(cls, dists):
        """
        Convolve a list of 1d float arrays together, using FFTs.
        The arrays need not have the same length, but each array should
        have length at least 1.
        """

        result_length = 1 + sum((len(dist) - 1) for dist in dists)

        # Copy each array into a 2d array of the appropriate shape.
        rows = zeros((len(dists), result_length))
        for i, dist in enumerate(dists):
            rows[i, :len(dist)] = dist.values

        # Transform, take the product, and do the inverse transform
        # to get the convolution.
        fft_of_rows = fft.fft(rows)
        fft_of_convolution = fft_of_rows.prod(axis=0)
        convolution = fft.ifft(fft_of_convolution)

        # Assuming real inputs, the imaginary part of the output can
        # be ignored.
        return PMF(convolution.real)

    @classmethod
    def flatten(cls, dists):
        """
        Sum a set of distributions to produce a new distribution.
        """
        flat_dist = [0.0] * (max([len(dist) for dist in dists]))
        for dist in dists:
            for i, value in enumerate(dist.values):
                flat_dist[i] += value
        return PMF(flat_dist)

    @classmethod
    def match_sizes(cls, dists):
        """
        Map a set of PMFs to the same size
        """
        largest = max([len(x) for x in dists])
        return [x.expand_to(largest) for x in dists]

    @classmethod
    def max_of_two(cls, dist1, dist2):
        """
        Compute the PMF for the max of two PMF
        """
        dist1, dist2 = cls.match_sizes([dist1, dist2])
        new_dist = []
        for value in range(len(dist1)):
            prob = dist1.get(value) * dist2.get(value)
            prob += dist1.values[value] * sum(dist2.values[:value])
            prob += dist2.values[value] * sum(dist1.values[:value])
            new_dist.append(prob)
        return PMF(new_dist)

class PMFCollection(object):
    """
    Discrete Probability Distribution - Used to keep track of collections of PMFs
    """
    def __init__(self, pmfs=None):
        self.pmfs = pmfs if pmfs is not None else []

    def __bool__(self):
        return len(self) > 0

    def __len__(self):
        return len(self.pmfs)

    def __str__(self):
        return str([x.values for x in self.pmfs])

    def get(self, index, defualt=None):
        try:
            return self.pmfs[index]
        except IndexError:
            return defualt

    def thresh_mod(self, thresh_mod):
        """
        Modify the collection based on a dice modifer
        """
        if thresh_mod == 0 or self.pmfs == []:
            return self
        elif thresh_mod > 0:
            return PMFCollection(self.pmfs[:1] * thresh_mod + self.pmfs)
        elif thresh_mod < 0:
            return PMFCollection((self.pmfs + self.pmfs[-1:] * (-1 * thresh_mod))[-1*len(self.pmfs):])

    def mul_col(self, col):
        return PMFCollection([self.mul_pmf(pmf) for pmf in col.pmfs])

    def mul_pmf(self, pmf):
        """
        Multiply the collection with a pmf and return a convolution of the results
        """
        new_pmfs = []
        for i, value in enumerate(pmf.values):
            new_pmfs.append(self.get(i, PMF.static(0)) * value)
        return PMF.flatten(new_pmfs)

    def convolve(self):
        """
        Convolve the PMFs into a single PMF
        """
        return PMF.convolve_many(self.pmfs)

    def convert_binomial_less_than(self, thresh):
        """
        Convert the PMFs into a Binomial distribution for values less than the thresh
        """
        return self.map(lambda x: x.convert_binomial_less_than(thresh))

    def convert_binomial(self, thresh):
        """
        Convert the PMFs into a Binomial distribution for values greater than or equal to the thresh
        """
        return self.map(lambda x: x.convert_binomial(thresh))

    def map(self, func):
        """
        Map a function over the PMFs in the PMFCollection
        """
        return PMFCollection([func(x) for x in self.pmfs])

    @classmethod
    def add_many(cls, collection_list):
        new_pmfs = []
        for i in range(max([len(x) for x in collection_list] or [0])):
            new_pmfs.append(PMF.convolve_many([x.get(i, PMF.static(0)) for x in collection_list if x]))
        return PMFCollection(new_pmfs)

    @classmethod
    def empty(cls):
        """
        Return an empty PMFCollection
        """
        return PMFCollection([])

    @classmethod
    def mdn(cls, m, n):
        """
        Return a PMFCollection of m dn PMFs
        """
        return PMFCollection([PMF.dn(n)] * m)

    @classmethod
    def static(cls, n):
        """
        Return a PMFCollection of a single static value
        """
        return PMFCollection([PMF.static(n)])





from .pmf import PMF
from .modifiers import ModifierCollection

class Weapon(object):
    """
    Holds the params of a weapon
    """
    def __init__(self, bs, shots, strength, ap, damage, modifiers=None):
        self.bs = bs
        self.shots = shots
        self.strength = strength
        self.ap = ap
        self.damage = damage
        self.modifiers = modifiers or ModifierCollection()